# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-20 02:38
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('django_remote_submission', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='uuid',
            field=models.UUIDField(default=uuid.uuid4, editable=False, help_text='A unique identifier for use in grouping Result files', verbose_name='Job UUID'),
        ),
    ]
